/*
FILE:     ISTable.C
*/
/*
VERSION:  8.004
*/
/*
DATE:     5/10/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/*!
** \file ISTable.C
**
** \brief Implementation file for ISTable class.
*/


#include <stdexcept>
#include <sstream>
#include <iostream>
#include <iomanip>

#include "Exceptions.h"
#include "GenString.h"
#include "ISTable.h"


const string ISTable::_version("V7");

const unsigned char ISTable::DEFAULT_OPTIONS = (ISTable::DT_STRING_VAL << 4);


ISTable::ISTable(const StringCompare::eCompareType colCaseSense) :
  _ttable(), _orient(eCOLUMN_WISE), _colCaseSense(colCaseSense)
{

    Init();

}


ISTable::ISTable(eOrientation orient, const StringCompare::eCompareType
  colCaseSense) : _ttable(), _orient(orient), _colCaseSense(colCaseSense)
{

    Init();

}


ISTable::ISTable(const string& name,
  const StringCompare::eCompareType colCaseSense) : _name(name),
  _ttable(), _orient(eCOLUMN_WISE), _colCaseSense(colCaseSense)
{

    Init();

}


ISTable::ISTable(const string& name, eOrientation orient,
  const StringCompare::eCompareType colCaseSense) : _name(name), _ttable(),
  _orient(orient), _colCaseSense(colCaseSense)
{

    Init();

}


ISTable::ISTable(const ISTable& inTable)
{

    _name = inTable._name;

    _ttable = inTable._ttable;

    _orient = inTable._orient;

    _colCaseSense = inTable._colCaseSense;

    _colNames = inTable._colNames;

    _precision = inTable._precision;
    _compare_opts = inTable._compare_opts;

    _indexNames = inTable._indexNames;
    _listsOfColumns = inTable._listsOfColumns;
    _unique = inTable._unique;

    _fnav = inTable._fnav;

    _modified = inTable._modified;

    _indices = inTable._indices;

}


ISTable::~ISTable()
{

    Clear();

}


ISTable& ISTable::operator=(const ISTable& inTable)
{

    if (this != &inTable)
    {
        Clear();

        _name = inTable._name;

        _ttable = inTable._ttable;

        _orient = inTable._orient;

        _colCaseSense = inTable._colCaseSense;

        _colNames = inTable._colNames;

        _precision = inTable._precision;
        _compare_opts = inTable._compare_opts;

        _indexNames = inTable._indexNames;
        _listsOfColumns = inTable._listsOfColumns;
        _unique = inTable._unique;

        _fnav = inTable._fnav;

        _modified = inTable._modified;

        _indices = inTable._indices;
    }

    return(*this);

}


ISTable::eTableDiff ISTable::operator==(ISTable& inTable)
{

    if (_colCaseSense != inTable._colCaseSense)
        return(ISTable::eCASE_SENSE);

    if (_colNames.size() > inTable._colNames.size())
        return(ISTable::eMORE_COLS);

    if (_colNames.size() < inTable._colNames.size())
        return(ISTable::eLESS_COLS);

    if (_colNames != inTable._colNames)
        return(ISTable::eCOL_NAMES);

    if (GetNumRows() > inTable.GetNumRows())
        return(ISTable::eMORE_ROWS);

    if (GetNumRows() < inTable.GetNumRows())
        return(ISTable::eLESS_ROWS);

    vector<string> firstTableCol, secondTableCol;

    for (unsigned int colI = 0; colI < _colNames.size(); ++colI)
    {
        GetColumn(firstTableCol, _colNames[colI]);
        inTable.GetColumn(secondTableCol, _colNames[colI]);
        if (firstTableCol != secondTableCol)
            return(ISTable::eCELLS);
    }

    return(ISTable::eNONE);

}


void ISTable::Rename(const string& name)
{

    _name = name;

}


void ISTable::GetColumnNames(vector<string>& colNames) const
{

    colNames = _colNames;

}


bool ISTable::IsColumnPresent(const string& colName)
{

    try
    {
        GetColumnIndex(colName);

        return(true);
    }
    catch (NotFoundException)
    {
        return(false);
    }

}


void ISTable::AddColumn(const string& colName, const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "ISTable::AddColumn");
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows()))
    {
        throw out_of_range("Column size greater than number of rows in "\
          " ISTable::AddColumn");
    }

    InsertColumn(colName, _colNames.size(), col);

}


void ISTable::InsertColumn(const string& colName, const string& atColName,
  const vector<string>& col)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(atColName);

        InsertColumn(colName, colIndex, col);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::InsertColumn");
        throw;
    }

}


void ISTable::FillColumn(const string& colName, const vector<string>& col)
{

    if (col.empty())
    {
        return;
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows()))
    {
        throw out_of_range("Column size greater than number of rows in "\
          " ISTable::FillColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        unsigned int numColumns;

        if (_orient == eCOLUMN_WISE)
            numColumns = _ttable.GetNumColumns();
        else
            numColumns = _ttable.GetNumRows();

        if (colIndex < numColumns)
        {
            if (_orient == eCOLUMN_WISE)
            {
                _ttable.FillColumn(colIndex, col);
            }
            else
            {
                _ttable.FillRow(colIndex, col);
            }
        }
        else
        {
            if (_orient == eCOLUMN_WISE)
                _ttable.InsertColumn(colIndex, col);
            else
                _ttable.InsertRow(colIndex, col);
        }

        // Clears all indices build on the column colIndex
        // VLAD: Why do we clear them. Why not rebuilding them?
        for (unsigned int i = 0; i < _indices.size(); i++)
        {
            if (IsColumnInIndex(i, colIndex))
                _indices[i].clear();
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::FillColumn");
        throw;
    }

}


void ISTable::AppendToColumn(const string& colName, const string& cell)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        // VLAD: INDEX CORRUPTION
        // This is not safe after index/indices is/are created
        // indices will be corrupted

        unsigned int numColumns;
        if (_orient == eCOLUMN_WISE)
            numColumns = _ttable.GetNumColumns();
        else
            numColumns = _ttable.GetNumRows();

        if (colIndex < numColumns)
        {
            if (_orient == eCOLUMN_WISE)
            {
                _ttable.AddRow();
                _ttable(GetNumRows() - 1, colIndex) = cell;
            }
            else
            {
                _ttable.AddColumn();
                _ttable(colIndex, GetNumRows() - 1) = cell;
            }

        }
        else
        {
            vector<string> tmpCol;
            tmpCol.push_back(cell);

            if (_orient == eCOLUMN_WISE)
                _ttable.InsertColumn(colIndex, tmpCol);
            else
            {
                _ttable.InsertRow(colIndex, tmpCol);
            }
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::AppendToColumn");
        throw;
    }

}


void ISTable::AppendToColumn(const string& colName, const vector<string>& col)
{

    // VLAD: INDEX CORRUPTION
    // This is not safe after index/indices is/are created
    // indices will be corrupted

    unsigned int colIndex = GetColumnIndex(colName);

    AppendToColumn(colIndex, col);

}


void ISTable::GetColumn(vector<string>& col, const string& colName)
{

    col.clear();

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (_orient == eCOLUMN_WISE)
            _ttable.GetColumn(col, colIndex, 0, _ttable.GetNumRows());
        else
            _ttable.GetRow(col, colIndex, 0, _ttable.GetNumColumns());
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetColumn");
        throw;
    }

}


void ISTable::GetColumn(vector<string>& col, const string& colName,
  const unsigned int fromRowIndex, unsigned int toRowIndex)
{

    col.clear();

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        for (unsigned int rowI = fromRowIndex; rowI <= toRowIndex; ++rowI)
        {
            col.push_back(operator()(rowI, colIndex));
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetColumn");
        throw;
    }

}


void ISTable::GetColumn(vector<string>& col, const string& colName,
  const vector<unsigned int>& rowIndex)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        col.clear();

        for (unsigned int index = 0; index < rowIndex.size(); ++index)
        {
            col.push_back(operator()(rowIndex[index], colIndex));
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetColumn");
        throw;
    }

}


void ISTable::RenameColumn(const string& oldColName, const string& newColName)
{

    if (oldColName.empty())
    {
        // Empty old column name.
        throw EmptyValueException("Empty old column name",
          "STable::RenameColumn");
    }

    if (newColName.empty())
    {
        // Empty new column name.
        throw EmptyValueException("Empty new column name",
          "STable::RenameColumn");
    }

    if (IsColumnPresent(newColName))
    {
        // Column already exists
        throw AlreadyExistsException("Duplicate column name",
          "STable::RenameColumn");
    }

    try
    {
        unsigned int colIndex = GetColumnIndex(oldColName);

        _colNames[colIndex] = newColName;
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Old column not found",
          "STable::RenameColumn");
        throw;
    }

}


void ISTable::ClearColumn(const string& colName)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (_orient == eCOLUMN_WISE)
            _ttable.ClearColumn(colIndex);
        else
            _ttable.ClearRow(colIndex);

        // Clear all indices built on the column colIndex
        for (unsigned int i = 0; i < _indices.size(); i++)
        {
            if (IsColumnInIndex(i, colIndex))
                _indices[i].clear();
        }
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::ClearColumn");
        throw;
    }
}


void ISTable::DeleteColumn(const string& colName)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (_orient == eCOLUMN_WISE)
            _ttable.DeleteColumn(colIndex);
        else
            _ttable.DeleteRow(colIndex);

        _precision.erase(_precision.begin() + colIndex);
        _compare_opts.erase(_compare_opts.begin() + colIndex);

        // Deletes all indices build on the column colIndex
        for (unsigned int i = 0; i < _indices.size(); i++)
        {
            if (IsColumnInIndex(i, colIndex))
            {
                DeleteIndex(i);
            }
        }

        UpdateColListOnColDelete(colIndex);

        _colNames.erase(_colNames.begin() + colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::DeleteColumn");
        throw;
    }

}


unsigned int ISTable::AddRow(const vector<string>& row)
{

    return(InsertRow(GetNumRows(), row));

}


unsigned int ISTable::InsertRow(const unsigned int rowIndex,
  const vector<string>& row)
{

    if (_colNames.empty())
    {
        // Column not found.
        throw EmptyContainerException("No columns in table",
          "ISTable::InsertRow");
    }

    if (GetNumRows() == 0)
    {
        for (unsigned int colI = 0; colI < _colNames.size(); ++colI)
        {
           AppendToColumn(_colNames[colI], String::Empty);
        }
    }
    else
        if (_orient == eCOLUMN_WISE)
            _ttable.InsertRow(rowIndex);
        else
            _ttable.InsertColumn(rowIndex);

    FillRow(rowIndex, row);

    // VLAD: SEE IF WE NEED TO UPDATE INDICES FOR ADD ROW
    // RebuildIndices() or UpdateIndices(rowIndex);
    RebuildIndices();

    return(GetNumRows());

}


void ISTable::FillRow(const unsigned int rowIndex, const vector<string>& row)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::FillRow");
    }

    if (_orient == eCOLUMN_WISE)
        _ttable.FillRow(rowIndex, row);
    else
        _ttable.FillColumn(rowIndex, row);

    UpdateIndices(rowIndex);

}


void ISTable::GetRow(vector<string>& row, const unsigned int rowIndex,
  const string& fromColName, const string& toColName)
{

    row.clear();

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::GetRow");
    }

    try
    {
        unsigned int fromColIndex = 0;
        if (!fromColName.empty())
        {
            fromColIndex = GetColumnIndex(fromColName);
        }

        unsigned int toColIndex = GetNumColumns() - 1;
        if (!toColName.empty())
        {
            toColIndex = GetColumnIndex(toColName);
        }

        if (_orient == eCOLUMN_WISE)
            _ttable.GetRow(row, rowIndex, fromColIndex, toColIndex + 1);
        else
            _ttable.GetColumn(row, rowIndex, fromColIndex, toColIndex + 1);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetSubRow");
        throw;
    }

}


void ISTable::ClearRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::ClearRow");
    }

    if (_orient == eCOLUMN_WISE)
        _ttable.ClearRow(rowIndex);
    else
        _ttable.ClearColumn(rowIndex);

    UpdateIndices(rowIndex);

}


void ISTable::DeleteRow(const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::DeleteRow");
    }

    if (_orient == eCOLUMN_WISE)
        _ttable.DeleteRow(rowIndex);
    else
        _ttable.DeleteColumn(rowIndex);

    // VLAD - THINK OF MORE EFFICIENT INDEX RE-BUILDING IF POSSIBLE
    RebuildIndices();

}


void ISTable::DeleteRows(const vector<unsigned int>& rows)
{

    // VLAD ERROR CHECKING ALL THE INDICES in rows
    // VLAD IMPROVE WITH REVERSE ITERATORS or more meaningfull index loops

    for (unsigned int rowIndex = 0; rowIndex < rows.size(); ++rowIndex)
    {
        DeleteRow(rows[rows.size() - 1 - rowIndex]);
    }

}


void ISTable::FindDuplicateRows(vector<pair<unsigned int, unsigned int> >& duplRows, const vector<string>& colNames, const bool keep,
  const eSearchDir searchDir)
{

    duplRows.clear();

    vector<unsigned int> colIndices;

    int tableIndex;

    // VLAD: MAKE AND UTILIZE METHOD THAT CONVERTS FROM COLUMN NAMES
    // TO COLUMN INDICES
    // VLAD VALUE replace with GetColumnIndices()
    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        tableIndex = GetColumnIndex(colNames[i]);
        if (tableIndex >= 0)
  	    colIndices.push_back(tableIndex);
        else
        {
            // Column not found.
            throw NotFoundException("Column not found",
              "ISTable::FindDuplicateRows");
        }
    }

    FindDuplicateRows(colIndices, duplRows, keep, searchDir);

}


void ISTable::UpdateCell(const unsigned int rowIndex, const string& colName,
  const string& cell)
{

    try
    {
        unsigned int colIndex = GetColumnIndex(colName);

        if (rowIndex >= GetNumRows())
        {
            // Wrong row index.
            throw out_of_range("Invalid row index in ISTable::UpdateCell");
        }

        if (_orient == eCOLUMN_WISE)
            _ttable(rowIndex, colIndex) = cell;
        else
            _ttable(colIndex, rowIndex) = cell;

        UpdateColListOnCellUpdate(rowIndex, colIndex);
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::UpdateCell");
        throw;
    }
}


const string& ISTable::operator()(const unsigned int rowIndex,
      const string& colName) const
{

    try
    {
        if (rowIndex >= GetNumRows())
        {
            // Wrong row index.
            throw out_of_range("Invalid row index in ISTable::operator()");
        }

        unsigned int colIndex = GetColumnIndex(colName);

        if (_orient == eCOLUMN_WISE)
            return(_ttable(rowIndex, colIndex));
        else
            return(_ttable(colIndex, rowIndex));
    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::operator()");
        throw;
    }

}





void ISTable::FindDuplicateRows(const vector<unsigned int>& colIndices,
  vector<pair<unsigned int, unsigned int> >& duplRows, const unsigned int keep,
  const eSearchDir searchDir)
{

    duplRows.clear();

    // VLAD ERROR HANDLING - CHECK INDICES
    // for keep = 1, reports duplicate rows, but keeps them in table
    // for keep = 0, reports duplicate rows and deletes them from table
    // searchDir can be forward or backward

    unsigned int realRowIndex;
    tIndex::iterator pos;
    string value;

    StringCompare::eCompareType compareType = GetCompareType(colIndices);

    tIndex tmpindex(compareType);

    for (unsigned int rowIndex = 0; rowIndex < GetNumRows(); rowIndex++)
    {
        if (searchDir == eFORWARD)
        {
            realRowIndex = rowIndex;
        }
        else
        {
            realRowIndex = GetNumRows() - rowIndex - 1;
        }

        value = AggregateRow(colIndices, realRowIndex);

        pos = tmpindex.find(value);
        if (pos != tmpindex.end())
        {
            duplRows.push_back(make_pair(realRowIndex, pos->second));
        }
        else
        {
            tIndex::value_type valType(value, realRowIndex);
            tmpindex.insert(valType);
        }

        value.clear();
    }

    if (keep == 0)
    {
        vector<unsigned int> duplRowIndices;
        for (unsigned int rowI = 0; rowI < duplRows.size(); rowI++)
        {
            duplRowIndices.push_back(duplRows[rowI].first);
        }

        DeleteRows(duplRowIndices);
    }

}



void ISTable::CreateIndex(const string& indexName,
  const vector<string>& colNames, const unsigned int unique)
{

    //   if unique flag is set to 0, table rows with duplicate values
    //     are allowed
    //   if unique flag is set to 1, table must not have duplicate values
    //     and rows with repeated values are deleted.

    vector<unsigned int> colIndices;

    GetColumnsIndices(colIndices, colNames);

    CreateIndex(indexName, colIndices, unique);

}


bool ISTable::AreListsOfColumnsValid(const vector<unsigned int>& colIndices)
{

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        if (colIndices[index] >= GetNumColumns())
        {
            return(false);
        }
    }

    return(true);

}


void ISTable::CreateIndex(const string& indexName,
  const vector<unsigned int>& colIndices, const unsigned int unique)
{

    if (colIndices.empty())
    {
        throw EmptyValueException("Empty column indices",
          "ISTable::CreateIndex");
    }

    VerifyColumnsIndices(colIndices);

    int indexIndex = FindIndex(indexName);
    if (indexIndex != -1)
    {
        // Index with the same name found. Just update it.
        ClearIndex(indexIndex);
        _listsOfColumns[indexIndex] = colIndices;
        _unique[indexIndex] = unique;
    }
    else
    {
        StringCompare::eCompareType compareType = GetCompareType(colIndices);

        // Create one or another type of multimap with/without
        // case-insensitivity
        tIndex index(compareType);

        _indexNames.push_back(indexName);
        _listsOfColumns.push_back(colIndices);
        _indices.push_back(index);
        _unique.push_back(unique);
    }

    for (unsigned int i = 0; i < GetNumRows(); i++)
        UpdateIndex(indexName, i);

}


void ISTable::UpdateIndex(const string& indexName, const unsigned int rowIndex)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
    {
        // Search index not found.
        throw NotFoundException("Search index not found",
          "ISTable::UpdateIndex");
    }

    UpdateIndex(indexIndex, rowIndex);

}


void ISTable::RebuildIndex(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
    {
        // Search index not found.
        throw NotFoundException("Search index not found",
          "ISTable::RebuildIndex");
    }

    RebuildIndex(indexIndex);

}


void ISTable::DeleteIndex(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex == -1)
    {
        // Search index not found.
        throw NotFoundException("Search index not found",
          "ISTable::DeleteIndex");
    }

    DeleteIndex(indexIndex);

}


void ISTable::RebuildIndices()
{

    for (unsigned int i = 0; i < _indices.size(); i++)
        RebuildIndex(i);

}


void ISTable::CreateKey(const vector<string>& colNames)
{

    vector<unsigned int> colIndices;

    GetColumnsIndices(colIndices, colNames);

    CreateKey(colIndices);

}


void ISTable::CreateKey(const vector<unsigned int>& colIndices)
{

    // VLAD HARDCODED CONST
    CreateIndex("__key__", colIndices, 1);

}


void ISTable::SetFlags(const string& colName, const unsigned char newOpts)
{

    try {

    unsigned int colIndex = GetColumnIndex(colName);

    if (newOpts & DT_MASK)
    {
        _compare_opts[colIndex] &= 15;
        _compare_opts[colIndex] |= ((~0) & ( newOpts & DT_MASK ));
    }

    if (newOpts & CASE_INSENSE)
    {
        _compare_opts[colIndex] |= CASE_INSENSE;
    }
    else
        _compare_opts[colIndex] &= ~CASE_INSENSE;

    if (newOpts & W_SPACE_INSENSE)
        _compare_opts[colIndex] |= W_SPACE_INSENSE;
    else
        _compare_opts[colIndex] &= ~W_SPACE_INSENSE;

    ValidateOptions(colIndex);

    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::SetFlags");
        throw;
    }
}


unsigned char ISTable::GetDataType(const string& colName)
{

    try {

    unsigned int colIndex = GetColumnIndex(colName);

    return((_compare_opts[colIndex] & DT_MASK) >> 4);

    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetDataType");
        throw;
    }
}


void ISTable::UpdateIndex(const unsigned int indexIndex,
  const unsigned int rowIndex)
{

    // indexIndex is the index of an index
    // rowIndex is the row index
    // This function preforms the following:
    //   if unique flag is set to 0, table rows with duplicate values
    //     are allowed
    //   if unique flag is set to 1, table must not have duplicate values
    //     and rows with repeated values are deleted.
    //
    // More information on unique flag being set to 1. If a row with rowIndex
    // already exists and the value of the new row with rowIndex already
    // exists, delete the row with rowIndex from the table and from the index.
    // VLAD: SIMPLIFY THE ABOVE unique == 1 processing.

    // VLAD ERROR HANDLING
    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::UpdateIndex");
    }

    if (!AreListsOfColumnsValid(_listsOfColumns[indexIndex]))
    {
        // VLAD - ERROR HANDLING - A LOT OF PLACES ATTEMPT TO CREATE INDICES
        // ON EMPTY COLUMNS
        return;
    }

    string value = SubRowValue(_listsOfColumns[indexIndex], rowIndex);

    if (_unique[indexIndex] == 0)
    {
        // Duplicate values are allowed, but not duplicate row indices.

        // VLAD IMPROVE. This and below for loop can be made into a method
        // IndexSearchForRow()

        // See if a pair with the row having rowIndex already exists. If it
        // does delete it first.
        for (tIndex::iterator pos = _indices[indexIndex].begin();
          pos != _indices[indexIndex].end(); ++pos)
        {
            if (pos->second == rowIndex)
            {
                _indices[indexIndex].erase(pos);
                break;
            }
        }

        // Duplicate values are allowed, just store value/index pair.
        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);

        return;
    }

    // Assuming that _unique[indexIndex] is 1
    pair<tIndex::iterator, tIndex::iterator> range;

    range = _indices[indexIndex].equal_range(value);

    if (range.first == range.second)
    {
        // Nothing found. Just store value/index pair.
        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);
    }
    else
    {
        // Found a value. If a row with rowIndex already exists,
        // remove it from the indices and delete it from the table.
        for (tIndex::iterator pos = _indices[indexIndex].begin();
          pos != _indices[indexIndex].end(); ++pos)
        {
            if (pos->second == rowIndex)
            {
                _indices[indexIndex].erase(pos);
                DeleteRow(rowIndex);
                break;
            }
        }
    }

}


void ISTable::RebuildIndex(const unsigned int indexIndex)
{

    if (!AreListsOfColumnsValid(_listsOfColumns[indexIndex]))
    {
        // VLAD - ERROR HANDLING - A LOT OF PLACES ATTEMPT TO CREATE INDICES
        // ON EMPTY COLUMNS
        return;
    }

    _indices[indexIndex].clear();

    string value;

    for (unsigned int rowIndex = 0; rowIndex < GetNumRows(); rowIndex++)
    {
        value = SubRowValue(_listsOfColumns[indexIndex], rowIndex);

        tIndex::value_type valType(value, rowIndex);
        _indices[indexIndex].insert(valType);

        value.clear();
    }

}


void ISTable::ClearIndex(const unsigned int indexIndex)
{

    _listsOfColumns[indexIndex].clear();
    _indices[indexIndex].clear();

}


void ISTable::DeleteIndex(const unsigned int indexIndex)
{

    // VLAD: THIS MAY NOT BE NEEDED. DEPENDS ON HOW STL CONTAINERS
    // FREE RESOURCES IN ERASE CALLS
    ClearIndex(indexIndex);

    _indexNames.erase(_indexNames.begin() + indexIndex);
    _listsOfColumns.erase(_listsOfColumns.begin() + indexIndex);
    _indices.erase(_indices.begin() + indexIndex);
    _unique.erase(_unique.begin() + indexIndex);

}


int ISTable::FindIndex(const string& indexName)
{

    // VLAD: FOR LOOP SEARCH: Use STL construct given below for this search
    // int index = distance(strings.begin(), find(strings.begin(),
    // strings.end(), string("xyz")));
    for (unsigned int i = 0; i < _indexNames.size(); i++)
    {
        if (_indexNames[i] == indexName)
        {
            return(i);
        }
    }

    return(-1);

}


int ISTable::FindIndex(const vector<unsigned int>& colIndices)
{

  int equal;

  if (colIndices.empty())
  {
      return(-1);
  }

  StringCompare::eCompareType compareType = GetCompareType(colIndices);

  for (unsigned int i = 0; i < _indices.size(); i++)
  {
      if (_listsOfColumns[i].size() != colIndices.size())
      {
          continue;
      }

      // VLAD INDEX SIMPLIFY CompareColIndices()
      // Assume that all columns are equal
      equal = 0;
      for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
      {
          if (_listsOfColumns[i][j] != colIndices[j])
          {
              // Found one non-equal column. Set the flag and brake the loop.
              equal = 1;
              break;
          }
      }

      if (equal == 0)
      {
          if ((_indices[i].key_comp()).GetCompareType() == compareType)
          {
              // The index is matched only if it also has the same
              // compare type
              return(i);
          }
      }
  }

  return(-1);

}


void ISTable::UpdateIndices(const unsigned int rowIndex)
{

    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        UpdateIndex(i, rowIndex);
    }

}


void ISTable::ClearIndices()
{

    for(unsigned int i = 0; i < _indexNames.size(); i++)
    {
        ClearIndex(i);
    }

}


bool ISTable::IsColumnInIndex(const unsigned int indexIndex,
  const unsigned int colIndex)
{

    vector<unsigned int>::iterator pos =
      find(_listsOfColumns[indexIndex].begin(),
      _listsOfColumns[indexIndex].end(), colIndex);

    if (pos != _listsOfColumns[indexIndex].end())
    {
        return(true);
    }
    else
    {
        return(false);
    }

}


int ISTable::FindKeyIndex()
{

    // VLAD HARDCODED CONST
    return(FindIndex("__key__"));

}


string ISTable::CellValue(const unsigned int colIndex,
  const unsigned int rowIndex)
{

    return(ConvertString(operator()(rowIndex, colIndex), colIndex));

}


string ISTable::ConvertString(const string& value,
  const unsigned int colIndex)
{
  string ret;
  unsigned char temp = _compare_opts[colIndex];
  switch (( temp & DT_MASK) >> 4) {
  case DT_INTEGER_VAL: {
      ConvertToInt(value, ret);
      break;
  }
#ifdef VLAD_DELETED
  case DT_DOUBLE_VAL: {
      ConvertDouble(value, ret);
      break;
  }
#endif
  case DT_STRING_VAL: {
      if ( temp & SC_MASK ) {
        if ( temp & WS_MASK ) {
          ConvertToLowerNoWhiteSpace(value, ret);
	  break;
	}
        else {
          String::LowerCase(value, ret);
	  break;
	}
      } else if (temp & WS_MASK) {
	String::RemoveWhiteSpace(value, ret);
	break;
      } else {
	ret=value;
	break;
      }
  }
  default: {
      ret=value;
      break;
  }
  }
  return ret;
}


string ISTable::MultiStringsValue(const vector<string>& values,
    const vector<unsigned int>& colIndices)
{

    string value;

    for (unsigned int index = 0; index < values.size(); index++)
    {
        AppendToAndDelimit(value,
          ConvertString(values[index], colIndices[index]));
    }

    return(value);

}


string ISTable::SubRowValue(const vector<unsigned int>& colIndices,
  const unsigned int rowIndex)
{

    string value;

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        AppendToAndDelimit(value, CellValue(colIndices[index], rowIndex));
    }

    return(value);

}


string ISTable::AggregateRow(const vector<unsigned int>& colIndices,
  const unsigned int rowIndex)
{

    string value;

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        AppendToAndDelimit(value, operator()(rowIndex, colIndices[index]));
    }

    return(value);

}


void ISTable::SetUnion(const vector<unsigned int>& a,
  const vector<unsigned int>& b, vector<unsigned int>& ret)
{

  ret.clear();

  unsigned int morea, moreb, ia, ib, lena, lenb, itema, itemb;

  ia = ib = morea = moreb = lena = lenb = 0;

  if (!a.empty())
  {
    lena  = a.size();
    morea = ia < lena;
  }

  if (morea)
    itema = a[ia];
  else
    itema = INT_MAX;

  if (!b.empty())
  {
    lenb  = b.size();
    moreb = ib < lenb;
  }

  if (moreb)
    itemb = b[ib];
  else
    itemb = INT_MAX;

  while (morea || moreb)
  {
    if (itema < itemb)
    {
      ret.push_back(itema);
      ia++; morea = ia < lena;
      if (morea)
        itema = a[ia];
      else
        itema = INT_MAX;
    }
    else if ( itema  == itemb)
    {
      ret.push_back(itema);

      ia++; morea = ia < lena;
      if (morea)
        itema = a[ia];
      else
        itema = INT_MAX;

      ib++; moreb = ib < lenb;
      if (moreb)
        itemb = b[ib];
      else
        itemb = INT_MAX;
    }
    else
    {
      ret.push_back(itemb);
      ib++; moreb = ib < lenb;
      if (moreb)
        itemb = b[ib];
      else
        itemb = INT_MAX;
    }
  }

}


void ISTable::SetIntersect(const vector<unsigned int>& a,
  const vector<unsigned int>& b, vector<unsigned int>& ret)
{

    ret.clear();

    int more, ia, ib, lena, lenb;


    //  cerr << "SetIntersect() Starting " << endl;

    if (a.empty() || b.empty())
        return;

    lena = a.size();
    lenb = b.size();

    ia = ib = 0;
    more = 1;

    while (more)
    {
        if (a[ia] < b[ib])
        {
            ia++;
            more = ia < lena;
        }
        else if (a[ia] == b[ib])
        {
            ret.push_back(a[ia]);
            ia++;
            ib++;
            more = (ib < lenb) && (ia < lena);
        }
        else
        {
            ib++;
            more = ib < lenb;
        }
    }

}


void ISTable::Init()
{

    _modified = false;
    _fnav = NULL;

}


void ISTable::Clear()
{

    _ttable.Clear();

    _colNames.clear();

    _modified = false;
    _indexNames.clear();
    _listsOfColumns.clear();
    _indices.clear();
    _unique.clear();
    _precision.clear();
    _compare_opts.clear();

}

unsigned int ISTable::GetColumnIndex(const string& colName) const
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::GetColumnIndex");
    }

    unsigned int colIndex;

    for (colIndex = 0; colIndex < _colNames.size(); ++colIndex)
    {
        if (String::IsEqual(_colNames[colIndex], colName, _colCaseSense))
        {
            break;
        }
    }

    if (colIndex == _colNames.size())
    {
        // Column not found.
        throw NotFoundException("Column not found",
          "ISTable::GetColumnIndex");
    }
    else
    {
        return(colIndex);
    }

}


void ISTable::GetColumnsIndices(vector<unsigned int>& colIndices,
  const vector<string>& colNames)
{

    if (colNames.empty())
    {
        _GetColumnsIndices(colIndices, _colNames);
    }
    else
    {
        _GetColumnsIndices(colIndices, colNames);
    }

}


void ISTable::_GetColumnsIndices(vector<unsigned int>& colIndices,
  const vector<string>& colNames)
{

    colIndices.clear();

    try
    {
        for (unsigned int index = 0; index < colNames.size(); index++)
        {
            colIndices.push_back(GetColumnIndex(colNames[index]));
        }
    }
    catch (RcsbException& rcsbException)
    {
        colIndices.clear();
        throw;
    }

}


void ISTable::ValidateOptions(unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in ISTable::ValidateOptions");
    }

    unsigned char a = 0, b = 0;

    b = _compare_opts[colIndex];

    a = b & DT_MASK;

    if (((a >> 4) > LAST_DT_VALUE) || (a == 0))
    {
        a = DEFAULT_OPTIONS;
    }
    else
        a = b;

    a &= ~(3 << 2); // not necessary, but clean up 3rd and 4th bit

    _compare_opts[colIndex] = a;

}


string ISTable::CreateInternalIndexName(const unsigned int indexIndex)
{

    // VLAD HARCODED CONST
    string name("index_");

    name += String::IntToString(indexIndex);

    return(name);

}


void ISTable::Search(vector<unsigned int>& res, const string& target,
  const string& colName, const eSearchType searchType)
{

    res.clear();

    unsigned int colIndex = GetColumnIndex(colName);

    Search(res, target, colIndex);

}


void ISTable::Search(vector<unsigned int>& res, const string& target,
  const unsigned int colIndex, const eSearchType searchType)
{

    res.clear();

    vector<string> targets;
    targets.push_back(target);

    vector<unsigned int> colIds;

    colIds.push_back(colIndex);

    Search(res, targets, colIds, searchType);

}


void ISTable::Search(vector<unsigned int>& res, const vector<string>& targets,
  const vector<string>& colNames, const eSearchType searchType,
  const string& indexName)
{

    res.clear();

    if (targets.size() != colNames.size())
    {
        throw out_of_range("colNames and targets have different size "\
          "in ISTable::Search");
    }

    vector<unsigned int> colIds;

    GetColumnsIndices(colIds, colNames);

    Search(res, targets, colIds, searchType, indexName);

}


void ISTable::Search(vector<unsigned int>& res, const vector<string>& targets,
  const vector<unsigned int>& colIds, const unsigned int indexNum,
  const eSearchType searchType)
{

    res.clear();

    int errCode;

    unsigned int len = targets.size();
    if ((len > colIds.size()) || (len < 1))
    {
        // To index less than from index.
        throw out_of_range("Invalid search target range in ISTable::Search");
    }

    VerifyColumnsIndices(colIds);

    string value = MultiStringsValue(targets, colIds);

    /*
    ** This code searches the container for all key/value pairs, where key
    ** equals "value" and stores all the values in the ret.
    */

    pair<tIndex::iterator, tIndex::iterator> range;

    range = _indices[indexNum].equal_range(value);

    tIndex::iterator start, afterLast;

    switch (searchType)
    {
        case eEQUAL:
           start = range.first;
           afterLast = range.second;
           break;
        case eLESS_THAN:
           start = _indices[indexNum].begin();
           afterLast = range.first;
           break;
        case eLESS_THAN_OR_EQUAL:
           start = _indices[indexNum].begin();
           afterLast = range.second;
           break;
        case eGREATER_THAN:
           start = range.second;
           afterLast = _indices[indexNum].end();
           break;
        case eGREATER_THAN_OR_EQUAL:
           start = range.first;
           afterLast = _indices[indexNum].end();
           break;
        default:
           // VLAD ERROR CHECKING: DO THIS CHECK PRIOR TO SEARCH
           errCode = COLUMN_OUT_OF_BOUNDS;
           return;
           break;
    }

    for (tIndex::iterator pos = start; pos != afterLast; ++pos)
    {
        res.push_back(pos->second);
    }

    if (searchType == eEQUAL)
        sort(res.begin(), res.end());

}


void ISTable::Search(vector<unsigned int>& res, const vector<string>& targets,
  const vector<string>& colNames, const unsigned int indexIndex,
  const eSearchType searchType)
{

    res.clear();

    vector<unsigned int> colIds;

    GetColumnsIndices(colIds, colNames);

    Search(res, targets, colIds, indexIndex, searchType);

}


void ISTable::Search(vector<unsigned int>& res, const vector<string>& targets,
  const vector<unsigned int>& colIds, const eSearchType searchType,
  const string& indexName)
{

    res.clear();

    int indexIndex;

    if (indexName.empty())
    {
        indexIndex = FindIndex(colIds);

        if (indexIndex == -1)
        {
            // No search index found. Create a new one.
            indexIndex = _indices.size();
            string name = CreateInternalIndexName(indexIndex);
            CreateIndex(name, colIds);
        }
    }
    else
    {
        indexIndex = FindIndex(indexName);
        if (indexIndex == -1)
        {
            // Search index not found.
            throw NotFoundException("Search index not found",
              "ISTable::Search");
        }
    }

    Search(res, targets, colIds, indexIndex, searchType);

}


unsigned int ISTable::FindFirst(const vector<string>& targets,
  const vector<string>& colNames, const string& indexName)
{

    if (targets.size() != colNames.size())
    {
        throw out_of_range("colNames and targets have different size "\
          "in ISTable::FindFirst");
    }

    vector<unsigned int> colIds;

    GetColumnsIndices(colIds, colNames);

    return(FindFirst(targets, colIds, indexName));

}


unsigned int ISTable::FindFirst(const vector<string>& targets,
  const vector<unsigned int>& colIds, const string& indexName)
{

    int indexIndex;

    if (indexName.empty())
    {
        indexIndex = FindIndex(colIds);
        if (indexIndex == -1)
        {
            indexIndex = _indices.size();
            string name = CreateInternalIndexName(indexIndex);
            CreateIndex(name, colIds);
        }
    }
    else
    {
        indexIndex = FindIndex(indexName);
        if (indexIndex == -1)
        {
            // Search index not found.
            throw NotFoundException("Search index not found",
              "ISTable::FindFirst");
        }
    }

    return(FindFirst(targets, colIds, indexIndex));

}


unsigned int ISTable::FindFirst(const vector<string>& targets,
  const vector<unsigned int>& colIds, const unsigned int indexNum)
{

    unsigned int ret = GetNumRows();

    if (colIds.empty())
    {
        throw EmptyValueException("Empty column indices",
          "ISTable::FindFirst");
    }

    unsigned int len = targets.size();
    if ((len < colIds.size()) || (len < 1))
    {
        // To index less than from index.
        throw out_of_range("Invalid search target range in ISTable::FindFirst");
    }

    VerifyColumnsIndices(colIds);

    string value = MultiStringsValue(targets, colIds);

    // Assume that it is not found

    tIndex::iterator pos = _indices[indexNum].find(value);
    if (pos != _indices[indexNum].end())
    {
        // Found
        ret = pos->second;
    }

    return(ret);

}


void ISTable::Merge(ISTable& inTable, unsigned int typeOfMerge) {
  unsigned int i,j;
  int index;
  vector<string> ColNames1;
  vector<string> ColNames2;
  vector<string> target;
  string cell;
  string empty("");

  GetColumnNames(ColNames1);
  inTable.GetColumnNames(ColNames2);


  vector<string> newColumn;
  // VLAD: EXAMINE IF CURRENT MERGE OPERATION WORKS AS IT SHOULD FROM THE
  // PERSPECTIVE OF _numDels. THIS IS WHERE THE LATEST BUG MAY BE
  // VLAD: GetNumRows()?
  for (i=0; i<GetNumRows(); i++)
    newColumn.push_back(empty);
  for (i=0; i<inTable._colNames.size(); i++) {
    if (!IsColumnPresent(ColNames2[i])) {
      AddColumn(ColNames2[i]);
      FillColumn(newColumn,_colNames.size()-1);
      // VLAD Do we need to EnlargeRowMap()?
    }
  }
  ColNames1.clear();
  GetColumnNames(ColNames1);
  vector<string> newRow;
  // VLAD: GetNumRows()?
  for (i=0; i<inTable.GetNumRows(); i++) {
    target.clear();
    for (j=0; j<_colNames.size(); j++) {
      cell.clear();
      if ((index = inTable.GetColumnIndex(ColNames1[j]))>=0) {
	cell = inTable(i, index);
	newRow.push_back(cell);
      }
      else
	newRow.push_back(empty);
    }

    // Find out what column ids are in key index
    int indexIndex = FindKeyIndex();
    for (unsigned int jj=0; jj< inTable._listsOfColumns[indexIndex].size();
      jj++) {
      cell.clear();
      cell = inTable(i, inTable._listsOfColumns[indexIndex][jj]);
      target.push_back(cell);
    }
    vector<unsigned int> result;
    Search(result, target, _listsOfColumns[indexIndex], indexIndex);
    if (result.empty()) {
      AddRow();
      FillRow(GetLastRowIndex(), newRow);
    }
    else {
      if (typeOfMerge == 1) {
	// overlap
	for (j=0; j<_colNames.size(); j++) {
	  cell.clear();
	  cell = operator()(result[0], j);
	  if (cell.empty()) {
	    index = inTable.GetColumnIndex(ColNames1[j]);
	    if (index>=0) {
	      cell.clear();
	      cell = inTable(i, (unsigned int)index);
	      UpdateCell(cell,j,result[0]);
	    }
	  }
	}
      }
      else {
	// overwrite
        DeleteRow(result[0]);
        InsertRow(result[0], newRow);
      }
    }
    target.clear();
    newRow.clear();
  }
  ColNames1.clear();
  ColNames2.clear();

}


ostream& operator<<(ostream& out, const ISTable& isTable)
{

    vector<string> colNames;

    isTable.GetColumnNames(colNames);

    unsigned int numRows = isTable.GetNumRows();

    out << endl << "Table " << isTable.GetName() <<  "  Columns: " <<
      colNames.size() << "  Rows: " << numRows << endl << endl;

    out << setw(5) << "RowNo";

    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        // VLAD HARDCODED CONST
        out << setw(10) << colNames[i] << " ";
    }
    out << endl;

    for (unsigned int j = 0; j < numRows; j++)
    {
        out << setw(5) << j;

        for (unsigned int i = 0; i < colNames.size(); i++)
        {
            // VLAD HARDCODED CONST
            out << setw(10) << isTable(j, colNames[i]) << " ";
        }

        out << endl;
    }

    return(out);

}


bool ISTable::PrintDiff(ISTable& inTable)
{
  bool ret = true;
  unsigned int i;
  unsigned int j;
  int index;
  vector<string> ColNames1;
  vector<string> ColNames2;
  vector<string> target;
  vector<unsigned int> sameCol1, sameCol2;
  vector<unsigned int> diffRow1, diffRow2;
  vector<unsigned int> sameRow1, sameRow2;
  string cell;
  string cell1, cell2;
  string Name1;
  string Name2;
  ofstream rpt;


  Name1=GetName();
  Name2=inTable.GetName();

  GetColumnNames(ColNames1);
  inTable.GetColumnNames(ColNames2);

  cout<<"** Compares table "<<Name1<<" and table "<<Name2<<" **"<<endl;

  cout<<"----------------- Columns report -----------------"<<endl;

  for (i=0; i<_colNames.size(); i++) {
    if (!inTable.IsColumnPresent(ColNames1[i])) {
      cout.width(5);
      cout<<i<<"  < ("<<ColNames1[i]<<")"<<endl;
      ret = false;
    }
    else {
      index = inTable.GetColumnIndex(ColNames1[i]);
      sameCol1.push_back(GetColumnIndex(ColNames1[i]));
      sameCol2.push_back(index);
      cout.width(5);
      cout<<i;
      cout.width(20);
      cout<<index<<"    ("<<ColNames1[i]<<")"<<endl;
    }
  }

  for (i=0; i<(inTable._colNames.size()); i++) {
    if (!IsColumnPresent(ColNames2[i])) {
      cout.width(25);
      cout<<i<<"  < ("<<ColNames2[i]<<")"<<endl;
      ret = false;
    }
  }

  cout<<endl;
  cout<<"------------------ Rows report ------------------"<<endl;
  for (i=0; i<GetNumRows(); i++) {
    // Find out what column ids are in key index
    int indexIndex = inTable.FindKeyIndex();
    for (unsigned int jj=0; jj< _listsOfColumns[indexIndex].size();
      jj++) {
      cell.clear();
      cell = operator()(i, _listsOfColumns[indexIndex][jj]);
      target.push_back(cell);
    }
    vector<unsigned int> result;
    inTable.Search(result, target, inTable._listsOfColumns[indexIndex],
      indexIndex);
    if (result.empty()) {
      // value for key is different
      cout.width(5);
      cout<<i<<"  <"<<endl;
      ret = false;
    }
    else {
      // looking for other then key column for same key value
      j=0;
      cell1.clear();
      cell2.clear();
      cell1 = operator()(i, sameCol1[j]);
      cell2 = inTable(result[0], sameCol2[j]);
      while(cell1 == cell2 &&j<sameCol1.size()) {
	cell1.clear();
	cell2.clear();
	cell1 = operator()(i, sameCol1[j]);
	cell2 = inTable(result[0], sameCol2[j]);
	j++;
      }
      if (cell1 == cell2) {
	cout.width(5);
	cout<<i;
	cout.width(10);
	cout<<result[0]<<endl;
      }
      else {
	diffRow2.push_back(result[0]);
	cout.width(5);
	cout<<i<<"  *"<<endl;
	ret = false;
      }
    }
    target.clear();
  }
    for (i=0; i<diffRow2.size(); i++){
    cout.width(15);
    cout<<diffRow2[i]<<"  *"<<endl;
  }
  cout<<endl;
  ColNames1.clear();
  ColNames2.clear();

  return ret;
}


void ISTable::Print(unsigned int indexNum)
{
  unsigned int i;
  tIndex::iterator pos;
  // int j;

  cout << endl << "Table " << _name <<  "  Columns: " << _colNames.size() << "  Rows: " << GetNumRows() << endl << endl;

      cout << setw(5)<< "RowNo";

  for (i = 0; i < _colNames.size(); i++)
  {
    // VLAD HARDCODED CONST
    cout << setw(10) << _colNames[i] << " ";
  }
  cout << endl;

  for (pos = _indices[indexNum].begin(); pos != _indices[indexNum].end(); ++pos)
  {
    cout << setw(5) << pos->second;

    for (i = 0; i < _colNames.size(); i++)
    {
        cout << setw(10) << operator()(pos->second, i) << " ";
    }

    cout << endl;
  }

}


void ISTable::Print(const string& indexName)
{

    int indexIndex = FindIndex(indexName);
    if (indexIndex != -1)
        Print(indexIndex);

}


void ISTable::SetFileNavigator(FileNavigator* fileNavigator)
{

    _fnav = fileNavigator;

}


int ISTable::WriteObject(FileNavigator* fileNav, int& size)
{

    return(WriteObjectV7(fileNav, size));

}


int ISTable::WriteObjectV7(FileNavigator* fileNav, int& size)
{

    int err;
    Word ret = 0, place = 0;
    unsigned int num;

    if (fileNav == NULL)
        return ERROR_NO_FILE_NAVIGATOR;

    string version = _version;

    err = fileNav->WriteString(version, ret);
    if (err) fileNav->PrintError(err);

    err = fileNav->WriteString(_name, place);
    if (err) fileNav->PrintError(err);

    int unused = 0;
    _ttable.Write(fileNav, unused);

    err = fileNav->WriteUWord(_orient, place);
    if (err) fileNav->PrintError(err);

    err = fileNav->WriteUWord(_colCaseSense, place);
    if (err) fileNav->PrintError(err);

    err = fileNav->WriteUWord(_colNames.size(), place);
    if (err) fileNav->PrintError(err);

    if (_colNames.empty())
    {
        size = place - ret + 1;

        return ret;
    }

    err = fileNav->WriteStrings(_colNames, place);

    err = fileNav->WriteUWords(_precision, place);

    string optsToWrite(_colNames.size(), ' ');
    for (unsigned int j = 0; j < _colNames.size(); ++j)
    {
        optsToWrite[j] = _compare_opts[j];
    }
    err = fileNav->WriteString(optsToWrite, place);

    num = _indexNames.size();
    err = fileNav->WriteUWord(num, place);
    if (err) fileNav->PrintError(err);

    if (num != 0)
    {
        err = fileNav->WriteStrings(_indexNames, place);

        for (unsigned int l = 0; l < num; ++l)
        {
            err = fileNav->WriteUWords(_listsOfColumns[l], place);
        }

        err = fileNav->WriteUWords(_unique, place);
    }

    size = place + ret + 1;

    return ret;

}


void ISTable::Read(unsigned int indexInFile)
{
    GetObject(indexInFile, _fnav);
}


int ISTable::Write()
{
    int size;

    return(WriteObject(_fnav, size));
}


int ISTable::GetObject(Word index, FileNavigator* fileNav) {
/*
  There are several GetObject methods to support reading tables
  from binary files saved with some of previous verson
*/
  string firstString;
  int err;
  string::size_type verStringIndex;

  Clear();

  fileNav->GetString(firstString, index, err);
  index++;

  // Reverse search of a table name for a version string

  verStringIndex = firstString.rfind(" $$$3");
  if (verStringIndex != string::npos)
  {
    // This version found. Remove the version string from the table name.
    firstString.erase(verStringIndex);
    Rename(firstString);
    return GetObjectV3(index,fileNav);
  }
  else
  {
    verStringIndex = firstString.rfind(" $$$2");
    if (verStringIndex != string::npos)
    {
      // This version found. Remove the version string from the table name.
      firstString.erase(verStringIndex);
      Rename(firstString);
      return GetObjectV2(index,fileNav);
    }
    else
    {
      // Reverse search of a table name for a version string
      verStringIndex = firstString.rfind(" $$$1");
      if (verStringIndex != string::npos)
      {
        // This version found. Remove the version string from the table name
        firstString.erase(verStringIndex);
        Rename(firstString);
        return GetObjectV1_1(index,fileNav);
      }
      else
      {
        if (firstString == _version)
        {
            return GetObjectV7(index,fileNav);
        }
        else if (firstString == "V6")
        {
            return GetObjectV6(index,fileNav);
        }
        else
        {
            Rename(firstString);
            return GetObjectV1(index,fileNav);
        }
      }
    }
  }
}


int ISTable::GetObjectV7(Word index, FileNavigator* fileNav)
{

    if (fileNav == NULL)
        return ERROR_NO_FILE_NAVIGATOR;

    int err;
    Word ret = 0;
    uWord numI = 0;
    uWord num;

    fileNav->GetString(_name, index, err);
    index++;

    index = _ttable.Read(index, fileNav);

    _orient = (eOrientation)fileNav->GetUWord(index, err);
    index++;

    _colCaseSense = (StringCompare::eCompareType)fileNav->GetUWord(index, err);
    index++;

    unsigned int numColumns = fileNav->GetUWord(index, err);
    index++;

    if (numColumns == 0)
    {
        return ret;
    }

    fileNav->GetStrings(_colNames, index, err);
    index++;

    if (numColumns != _colNames.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    fileNav->GetUWords(_precision, index, err);
    index++;

    if (numColumns != _precision.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    string oToGet;
    fileNav->GetString(oToGet, index, err);
    index++;

    if (numColumns != oToGet.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    for (unsigned int j = 0; j < oToGet.size(); j++)
    {
        _compare_opts.push_back(oToGet[j]);
    }

    // Number of search indices
    num = fileNav->GetUWord(index, err);
    index++;

    if (num != 0)
    {
        vector<string> idxNamesToGet;
        fileNav->GetStrings(idxNamesToGet, index, err);
        index++;
        numI = idxNamesToGet.size();

        if (numI != num)
            return INTERNAL_INCONSISTENCY_ERROR;

        // This version of odb file ( $$$3) and ISTables (_version == 5)
        // does not take serialized indices from the *.odb files, but
        // re-generates them.
        vector<uWord> listToGet;
        for (unsigned int j = 0; j < num; j++)
        {
            fileNav->GetUWords(listToGet, index, err);
            index++;
            CreateIndex(idxNamesToGet[j], listToGet);
            listToGet.clear();
        }

        fileNav->GetUWords(_unique, index, err);
        index++;
    }

    return ret;

}


int ISTable::GetObjectV6(Word index, FileNavigator* fileNav)
{

    if (fileNav == NULL)
        return ERROR_NO_FILE_NAVIGATOR;

    int err;
    Word ret = 0;
    uWord numI = 0;
    uWord num;

    fileNav->GetString(_name, index, err);
    index++;

    unsigned int numColumns = fileNav->GetUWord(index, err);
    index++;

    if (numColumns == 0)
    {
        return ret;
    }

    // unsigned int numRows = fileNav->GetUWord(index, err);
    // This is not used and is not read, only skipped
    index++;

    fileNav->GetStrings(_colNames, index, err);
    index++;

    if (numColumns != _colNames.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    fileNav->GetUWords(_precision, index, err);
    index++;

    if (numColumns != _precision.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    string oToGet;
    fileNav->GetString(oToGet, index, err);
    index++;

    if (numColumns != oToGet.size())
        return INTERNAL_INCONSISTENCY_ERROR;

    for (unsigned int j = 0; j < oToGet.size(); j++)
    {
        _compare_opts.push_back(oToGet[j]);
    }

    for (unsigned int k = 0; k < _colNames.size(); ++k)
    {
        vector<string> tmpCol;
        fileNav->GetStrings(tmpCol, index, err);
        index++;
        if (_orient == eCOLUMN_WISE)
           _ttable.InsertColumn(k, tmpCol);
        else
           _ttable.InsertRow(k, tmpCol);
    }

    // EnlargeRowMap(numRows);

    // Number of search indices
    num = fileNav->GetUWord(index, err);
    index++;

    if (num != 0)
    {
        vector<string> idxNamesToGet;
        fileNav->GetStrings(idxNamesToGet, index, err);
        index++;
        numI = idxNamesToGet.size();

        if (numI != num)
            return INTERNAL_INCONSISTENCY_ERROR;

        // This version of odb file ( $$$3) and ISTables (_version == 5)
        // does not take serialized indices from the *.odb files, but
        // re-generates them.
        vector<uWord> listToGet;
        for (unsigned int j = 0; j < num; j++)
        {
            fileNav->GetUWords(listToGet, index, err);
            index++;
            CreateIndex(idxNamesToGet[j], listToGet);
            listToGet.clear();
        }

        fileNav->GetUWords(_unique, index, err);
        index++;
    }

    return ret;

}


int ISTable::GetObjectV3(Word index, FileNavigator* fileNav)
{

  if (fileNav == NULL)
      return ERROR_NO_FILE_NAVIGATOR;

  int err;
  Word ret = 0;
  uWord numI = 0;
  uWord num;

  fileNav->GetWord(index, err);
  index++;

  unsigned int numColumns = fileNav->GetWord(index, err);
  index++;

  // unsigned int numRows = fileNav->GetWord(index, err);
  // This is not used and is not read, only skipped
  index++;

  fileNav->GetStrings(_colNames, index, err);
  index++;

  if (numColumns != _colNames.size())
      return INTERNAL_INCONSISTENCY_ERROR;

  fileNav->GetUWords(_precision, index, err);
  index++;

  string oToGet;
  fileNav->GetString(oToGet, index, err);
  index++;
  if (!_colNames.empty())
  {
    for (unsigned int j = 0; j < _colNames.size(); j++)
    {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  for (unsigned int k = 0; k < _colNames.size(); ++k)
  {
    vector<string> tmpCol;
    fileNav->GetStrings(tmpCol, index, err);
    index++;
    if (_orient == eCOLUMN_WISE)
        _ttable.InsertColumn(k, tmpCol);
    else
        _ttable.InsertRow(k, tmpCol);
  }

  // EnlargeRowMap(numRows);

  // Number of search indices
  num = fileNav->GetWord(index, err);
  index++;

  vector<string> idxNamesToGet;
  fileNav->GetStrings(idxNamesToGet, index, err);
  index++;
  numI = idxNamesToGet.size();

  if (numI != num)
      return INTERNAL_INCONSISTENCY_ERROR;

  fileNav->GetUWords(_unique, index, err);
  index++;

  // This version of odb file ( $$$3) and ISTables (_version == 5)
  // does not take serialized indices from the *.odb files, but
  // re-generates them.
  vector<uWord> listToGet;
  for (unsigned int j = 0; j < num; j++)
  {
    fileNav->GetUWords(listToGet, index, err);
    index++;
    CreateIndex(idxNamesToGet[j], listToGet);
    listToGet.clear();
  }

  return ret;

}


int ISTable::GetObjectV2(Word index, FileNavigator* fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;
  uWord numI = 0;
  uWord num;

  unsigned int version    = fileNav->GetWord(index, err); index++;

  if ((version==3) || (version==4))
  {
    num = fileNav->GetWord(index, err);index++;
    vector<Word> kToGet;
    fileNav->GetWords(index, kToGet, err);index++;
    kToGet.clear();
  }

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  unsigned int numRows    = fileNav->GetWord(index, err); index++;
  //EnlargeRowMap(numRows);
  int numDels = fileNav->GetWord(index, err); index++;

  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;

  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
      _colNames.push_back(temp);
    }
  }


  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  // VLAD: THIS CODE HAS NOT BEEN TESTED
  // NOTE: The data read here is a former ISTable member _deleted. Values
  // for _deleted used to be 0 or 1, where 0 indicated that row is not
  // deleted and 1 indicated that row was deleted.

  // Out of order read of _deleted, since it is needed for proper
  // construction of _data. Index is not advanced.
  vector<unsigned int> deleted;
  vector<uWord> dToGet;
  fileNav->GetUWords(dToGet, index + _colNames.size(), err);
  if (!dToGet.empty()) {
    for (j = 0; j < numRows; j++) {
      deleted.push_back(dToGet[j]);
    }
  }
  if (!dToGet.empty()) dToGet.clear();

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    vector<string> tmpCol;
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          if (!deleted[l])
              tmpCol.push_back(tempString[l]);
      }
      if (_orient == eCOLUMN_WISE)
          _ttable.InsertColumn(k, tmpCol);
      else
          _ttable.InsertRow(k, tmpCol);

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
          //free(tempColumns[l3]);
        }
        tempColumns.clear();
      }
      delete[] tempString;
    }
  }

  index++;

  num = fileNav->GetWord(index, err); index++;

  vector<string> idxNamesToGet;
  fileNav->GetStrings(idxNamesToGet, index, err);
  numI = idxNamesToGet.size();
  index++;

  if (numI != num) return INTERNAL_INCONSISTENCY_ERROR;
  if (numI > 0) {
    string  temp;
    for (j = 0; j < numI; j++) {
      if (idxNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = idxNamesToGet[j];
      }
        //_indexNames.push_back(temp);
    }
  }
  if (!idxNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < numI; l2++) {
      //free(idxNamesToGet[l2]);
    }
    // if (numI > 0) free(idxNamesToGet);
  }

  vector<uWord> uniqueToGet;
  fileNav->GetUWords(uniqueToGet, index, err);
  index++;
  if (!uniqueToGet.empty()) {
    for (j = 0; j < uniqueToGet.size(); j++) {
      _unique.push_back(uniqueToGet[j]);
    }
  }
  if (!uniqueToGet.empty()) uniqueToGet.clear();

  // This version of odb file ( $$$3) and ISTables (_version == 5)
  // does not take serialized indices from the *.odb files, but
  // re-generates them.
  vector<uWord> listToGet;
  for (j=0; j<num; j++)
  {
    fileNav->GetUWords(listToGet, index, err);
    index++;
    vector<unsigned int> list;
    for (unsigned int l2=0; l2<listToGet.size(); l2++){
      list.push_back(listToGet[l2]);
    }
    CreateIndex(idxNamesToGet[j], list);
    //_listsOfColumns.push_back(list);
    list.clear();
    if (!listToGet.empty()) listToGet.clear();
  }

   // Correct the members
  numRows -= numDels;

  // EnlargeRowMap(numRows);

  return ret;
}


int ISTable::GetObjectV1_1(Word index, FileNavigator* fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;

  fileNav->GetWord(index, err); index++;
  vector<Word> kToGet;
  fileNav->GetWords(index, kToGet, err);index++;
  kToGet.clear();

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  // unsigned int numRows    = fileNav->GetWord(index, err); index++;
  // This is not used and is not read, only skipped

  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;
  int treeAlloc  = (int) fileNav->GetWord(index, err); index++;
  unsigned int numTrees = fileNav->GetWord(index, err); index++;

  //EnlargeRowMap(numRows);

  vector<Word> cMap;
  fileNav->GetWords(index, cMap, err);index++;
  cMap.clear();


  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
        _colNames.push_back(temp);
    }
  }
  if (!colNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < _colNames.size(); l2++) {
      //free(colNamesToGet[l2]);
    }
    if (_colNames.size() > 0) colNamesToGet.clear();
  }

  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    vector<string> tmpCol;
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          tmpCol.push_back(tempString[l]);
      }
      if (_orient == eCOLUMN_WISE)
          _ttable.InsertColumn(k, tmpCol);
      else
          _ttable.InsertRow(k, tmpCol);

      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
          //free(tempColumns[l3]);
        }
        tempColumns.clear();
      }
      delete[] tempString;
    }
  }
/********************************************/
  int tableIndex = index ;
  if ((treeAlloc <= 0) || (numTrees <= 0)) {
    if (treeAlloc <= 0) treeAlloc = 0;
    if (numTrees <= 0) numTrees = 0;
    return INTERNAL_INCONSISTENCY_ERROR;
  }


  vector<Word> tInts;

  for (k = 0; k < numTrees; k++) {
    fileNav->GetWords(tableIndex + k, tInts, err); index++;
    tInts.clear();
  }

  return ret;
}



int ISTable::GetObjectV1(Word index, FileNavigator* fileNav) {
  if (!fileNav) return ERROR_NO_FILE_NAVIGATOR;
  int err;
  unsigned int j;
  Word ret = 0;
  uWord numC = 0;

  unsigned int numColumns = fileNav->GetWord(index, err); index++;
  // unsigned int numRows    = fileNav->GetWord(index, err); index++;
  // This is not used and is not read, only skipped

  // _colAlloc field is next. It is ignored since it is not used with
  // this version of ISTable
  index++;
  int treeAlloc  = (int) fileNav->GetWord(index, err); index++;
  unsigned int numTrees = fileNav->GetWord(index, err); index++;

  // EnlargeRowMap(numRows);

  vector<Word> cMap;
  fileNav->GetWords(index, cMap, err);index++;
  cMap.clear();


  vector<string> colNamesToGet;
  fileNav->GetStrings(colNamesToGet, index, err);
  numC = colNamesToGet.size();
  index++;

  if (numC != numColumns) return INTERNAL_INCONSISTENCY_ERROR;
  if (numC > 0) {
    string  temp;
    for (j = 0; j < numColumns; j++) {
      if (colNamesToGet[j].empty()) {
	  temp = "";
      } else {
	  temp = colNamesToGet[j];
      }
        _colNames.push_back(temp);
    }
  }
  if (!colNamesToGet.empty()) {
    for (unsigned int l2 = 0; l2 < _colNames.size(); l2++) {
      //free(colNamesToGet[l2]);
    }
    if (_colNames.size() > 0) colNamesToGet.clear();
  }

  vector<uWord> pToGet;
  fileNav->GetUWords(pToGet, index, err);
  index++;
  if (!pToGet.empty()) {
    for (j = 0; j < _colNames.size(); j++) {
      _precision.push_back(pToGet[j]);
    }
  }
  if (!pToGet.empty()) pToGet.clear();

  string oToGet;
  fileNav->GetString(oToGet, index, err); index++;
  if (_colNames.size() > 0) {
    for (j = 0; j < _colNames.size(); j++) {
      _compare_opts.push_back(oToGet[j]);
    }
  }

  unsigned int k, l;
  vector<string> tempColumns;
  for (k = 0; k < _colNames.size(); k++) {
    vector<string> tmpCol;
    fileNav->GetStrings(tempColumns, index, err);
    numC = tempColumns.size();
    index++;
    if (numC > 0) {
      string * tempString = new string[numC];
      for (l = 0; l < numC; l++) {
        if (tempColumns[l].empty()) {
          tempString[l] = "";
        }
        else {
          tempString[l] = tempColumns[l];
        }
          tmpCol.push_back(tempString[l]);
      }
      if (_orient == eCOLUMN_WISE)
          _ttable.InsertColumn(k, tmpCol);
      else
          _ttable.InsertRow(k, tmpCol);


      if (!tempColumns.empty()) {
        for (unsigned int l3 = 0; l3 < numC; l3++) {
         // free(tempColumns[l3]);
        }
        tempColumns.clear();
      }
      delete[] tempString;
    }
  }
/********************************************/
  int tableIndex = index ;
  if ((treeAlloc <= 0) || (numTrees <= 0)) {
    if (treeAlloc <= 0) treeAlloc = 0;
    if (numTrees <= 0) numTrees = 0;
    return INTERNAL_INCONSISTENCY_ERROR;
  }


  vector<Word> tInts;

  for (k = 0; k < numTrees; k++) {
    fileNav->GetWords(tableIndex + k, tInts, err); index++;
    tInts.clear();
  }

  return ret;
}


void ISTable::ConvertToInt(const string& a, string& ret)
{

  ret.clear();

  const unsigned int len = a.size();

  if (len > INT_LIMIT)
  {
    return;
  }

  int intValue = 0;

  if (len != 0)
  {
      intValue = String::StringToInt(a);
  }

  if (intValue < 0)
  {
      const string maxNum(len - 1, '9');

      ret.push_back('-');
      ret.append(INT_LIMIT - len, '9');

      // maxNum - (-intValue)
      ret += String::IntToString(String::StringToInt(maxNum) + intValue);
  }
  else
  {
      ret.assign(INT_LIMIT - len, '0');
      ret += a;
  }

}


void ISTable::ConvertDouble(const string& a, string& ret){
/* This method convert string representing a double value into
   another string. when we wont to compare two double value
   we can compare two string value (converted value) and
   result will be the same
*/

  const int MAX_DECIMAL_DIGIT = 9;
  string m;  // Mantissa before decimal point
  string r;  // Mantissa after decimal point (remainder)
  string e;  // Exponent, without letter "e"
  string aa; // Exponential representation of the input double
  // Maximum remainder complement value

  int eint;  // Exponent value
  double adouble; // Input value


  // Convert the original double representation to mantissa/exponent
  // representation.
  adouble = String::StringToDouble(a);

  ostringstream outStringStream;
  outStringStream << std::scientific << std::showpoint <<
    std::setprecision(MAX_PRECISION) << adouble;

  aa = outStringStream.str();

  // Extract mantissa, remainder and exponent
  string::size_type dotCharIndex = aa.find('.');
  if (dotCharIndex == string::npos)
  {
    return;
  }

  m.append(aa, 0, dotCharIndex);

  string::size_type expCharIndex = aa.find('e');
  if (expCharIndex == string::npos)
  {
    // VLAD BUG: Fix this
    return;
  }

  r.append(aa, dotCharIndex + 1, expCharIndex - dotCharIndex - 1);
  e.append(aa, expCharIndex + 1, aa.size() - expCharIndex - 1);

  eint = String::StringToInt(e);

  if (adouble < 0)
  {
    ret.push_back('-');
  }
  else
  {
    ret.push_back('0');
  }

  if (eint < 0)
  {
    const string maxExpNum(e.size() - 1, '9');

    ret.push_back('-');

    // Complement the exponent
    ret.append(EXPONENT - e.size(), '9');

    // maxExpNum - (-eint)
    ret += String::IntToString(String::StringToInt(maxExpNum) + eint);
  }
  else
  {
    ret.push_back('0');
    ret.append(EXPONENT - e.size(), '0');
    ret.append(e, 1, e.size() - 1);
  }

  if (adouble < 0)
  {
    // Complement mantissa
    // MAX_DECIMAL_DIGIT - (-mantissa before dot)
    ret += String::IntToString(MAX_DECIMAL_DIGIT + String::StringToInt(m));

    ret.push_back('.');

    // Complement each digit in the remainer
    for (unsigned int index = 0; index < r.size(); ++index)
    {
      ret += String::IntToString(MAX_DECIMAL_DIGIT -
        String::StringToInt(r.substr(index, 1)));
    }
  }
  else
  {
    ret += m;
    ret += '.';
    ret += r;
  }

}


void ISTable::ConvertToLowerNoWhiteSpace(const string& a, string& ret){
  string temp;
  String::RemoveWhiteSpace(a, temp);
  String::LowerCase(temp, ret);
}


StringCompare::eCompareType
  ISTable::GetCompareType(const vector<unsigned int>& colIndices)
{

    // Initially set it to case sensitive type.
    StringCompare::eCompareType compareType = StringCompare::eCASE_SENSITIVE;

    // In setting the compare type, only the options of the column
    // at index 0 are considered. It is assumed that all the rest of
    // the columns have the same options as the first column.

    unsigned char opts = _compare_opts[colIndices[0]];
    switch ((opts & DT_MASK) >> 4)
    {
        case DT_STRING_VAL:
            if (opts & SC_MASK)
            {
                // Case in-sensitive column.
                compareType = StringCompare::eCASE_INSENSITIVE;
            }
            break;
        default:
            break;
    }

    return(compareType);

}


void ISTable::UpdateColListOnColInsert(const unsigned int colIndex)
{

    // Updating lists of columns. If a column in the list has column
    // index >= colIndex then increment it by 1
    for (unsigned int i = 0; i < _listsOfColumns.size(); i++)
    {
        for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
        {
            if (_listsOfColumns[i][j] >= colIndex)
                _listsOfColumns[i][j]++;
        }
    }

}


void ISTable::UpdateColListOnColDelete(const unsigned int colIndex)
{

    // Updating lists of columns. If a column in the list has column
    // index >= colIndex then decrement it by 1
    for (unsigned int i = 0; i < _listsOfColumns.size(); i++)
    {
        for (unsigned int j = 0; j < _listsOfColumns[i].size(); j++)
        {
            if (_listsOfColumns[i][j] >= colIndex)
                _listsOfColumns[i][j]--;
        }
    }

}


void ISTable::UpdateColListOnCellUpdate(const unsigned int rowIndex,
  const unsigned int colIndex)
{

    for (unsigned int j = 0; j < _indices.size(); j++)
    {
        for (unsigned int k = 0; k < _listsOfColumns[j].size(); k++)
        {
            if (_listsOfColumns[j][k] == colIndex)
            {
                UpdateIndex(j, rowIndex);
                // VLAD: EXAMINE THIS !!
                break;
            }
        }
    }

}


void ISTable::FillColumn(const vector<string>& col,
  const unsigned int colIndex)
{

    if (col.empty())
    {
        return;
    }

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in ISTable::FillColumn");
    }

    for (unsigned int rowI = 0; rowI < col.size(); ++rowI)
    {
        if (_orient == eCOLUMN_WISE)
            _ttable(rowI, colIndex) = col[rowI];
        else
            _ttable(colIndex, rowI) = col[rowI];
    }

    // Clears all indices build on the column colIndex
    // VLAD: Why do we clear them. Why not rebuilding them?
    for (unsigned int i = 0; i < _indices.size(); i++)
    {
        if (IsColumnInIndex(i, colIndex))
            _indices[i].clear();
    }

}


void ISTable::AppendToColumn(const unsigned int colIndex,
  const vector<string>& col)
{

    // VLAD: INDEX CORRUPTION
    // This is not safe after index/indices is/are created
    // indices will be corrupted

    unsigned int numColumns;

    if (_orient == eCOLUMN_WISE)
        numColumns = _ttable.GetNumColumns();
    else
        numColumns = _ttable.GetNumRows();

    if (colIndex < numColumns)
    {
        unsigned int oldNumRows = GetNumRows();

        for (unsigned int colI = 0; colI < col.size(); ++colI)
        {
            if (_orient == eCOLUMN_WISE)
                _ttable.AddRow();
            else
                _ttable.AddColumn();
        }

        for (unsigned int colI = 0; colI < col.size(); ++colI)
        {
            if (_orient == eCOLUMN_WISE)
                _ttable(oldNumRows - 1 + colI, colIndex) = col[colI];
            else
                _ttable(colIndex, oldNumRows - 1 + colI) = col[colI];
        }
    }
    else
    {
        if (_orient == eCOLUMN_WISE)
            _ttable.InsertColumn(colIndex, col);
        else
            _ttable.InsertRow(colIndex, col);
    }

}


int ISTable::UpdateCell(const string& cell, const unsigned int colIndex,
  const unsigned int rowIndex)
{

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::UpdateCell");
    }

    if (_orient == eCOLUMN_WISE)
        _ttable(rowIndex,  colIndex) = cell;
    else
        _ttable(colIndex, rowIndex) = cell;

    UpdateColListOnCellUpdate(rowIndex, colIndex);

    return NO_TABLE_ERROR;

}


const string& ISTable::operator()(const unsigned int rowIndex,
  const unsigned int colIndex) const
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in ISTable::operator()");
    }

    if (rowIndex >= GetNumRows())
    {
        // Wrong row index.
        throw out_of_range("Invalid row index in ISTable::operator()");
    }

    if (_orient == eCOLUMN_WISE)
        return(_ttable(rowIndex, colIndex));
    else
        return(_ttable(colIndex, rowIndex));

}


int ISTable::SetFlags(const unsigned char newOpts, const unsigned int colIndex)
{

    if (colIndex >= _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in ISTable::SetFlags");
    }

    if (newOpts & DT_MASK)
    {
        _compare_opts[colIndex] &= 15;
        _compare_opts[colIndex] |= ((~0) & ( newOpts & DT_MASK ));
    }

    if (newOpts & CASE_INSENSE)
    {
        _compare_opts[colIndex] |= CASE_INSENSE;
    }
    else
        _compare_opts[colIndex] &= ~CASE_INSENSE;

    if (newOpts & W_SPACE_INSENSE)
        _compare_opts[colIndex] |= W_SPACE_INSENSE;
    else
        _compare_opts[colIndex] &= ~W_SPACE_INSENSE;

    ValidateOptions(colIndex);

    return NO_TABLE_ERROR;

}


void ISTable::CreateColumn(const string& colName, const unsigned int atColIndex,
  const vector<string>& col)
{

    if (colName.empty())
    {
        // Empty column name.
        throw EmptyValueException("Empty column name",
          "STable::CreateColumn");
    }

    if (atColIndex > _colNames.size())
    {
        // Wrong column index.
        throw out_of_range("Invalid column index in STable::CreateColumn");
    }

    if ((GetNumRows() != 0) && (col.size() > GetNumRows()))
    {
        // Wrong column index.
        throw out_of_range("Invalid column size in STable::FillColumn");
    }

    try
    {
        GetColumnIndex(colName);

        // Column already exists
        throw AlreadyExistsException("Duplicate column name",
          "STable::CreateColumn");
    }
    catch (NotFoundException)
    {
        // Continue execution, as it is normal that column is not found
        _colNames.insert(_colNames.begin() + atColIndex, colName);

        if (!col.empty())
        {
            if (_orient == eCOLUMN_WISE)
                _ttable.InsertColumn(atColIndex, col);
            else
                _ttable.InsertRow(atColIndex, col);
        }
    }

}


void ISTable::VerifyColumnsIndices(const vector<unsigned int>& colIndices)
{

    for (unsigned int index = 0; index < colIndices.size(); index++)
    {
        if (colIndices[index] >= _colNames.size())
        {
            // Wrong column index.
            throw out_of_range("Invalid column index in "\
              " ISTable::VerifyColumnsIndices");
        }
    }

}


void ISTable::InsertColumn(const string& colName,
  const unsigned int colIndex, const vector<string>& col)
{

    if (GetNumRows() != 0)
    {
        vector<string> newCol(GetNumRows(), String::Empty);
        CreateColumn(colName, colIndex, newCol);
    }
    else
    {
        CreateColumn(colName, colIndex, col);
    }

    // VLAD. It is better to introduce one data structure to
    // account for all column info. E.g. we could combine _compare_opts
    // and _precision to be two fields of a struct.
    _compare_opts.insert(_compare_opts.begin() + colIndex, DEFAULT_OPTIONS);

     // From ANSI C++ standard point of view, the re-casting of
     // DEFAULT_PRECISION to (unsigned int) is not necessary, since its type
     // is unsigned int. However, the linker issues unresolved symbol error
     // if this re-casting is omitted.
    _precision.insert(_precision.begin() + colIndex,
      (unsigned int)DEFAULT_PRECISION);

    SetFlags(DEFAULT_OPTIONS, colIndex);

    if (colIndex != (_colNames.size() - 1))
    {
        // If it is an insert, the column indices have to be updated.
        UpdateColListOnColInsert(colIndex);
    }

    FillColumn(col, colIndex);

}


void ISTable::GetColumn(vector<string>& col, const string& colName,
  const string& indexName)
{

    try {

    unsigned int colIndex = GetColumnIndex(colName);

    col.clear();

    unsigned int indexIndex = FindIndex(indexName);

    col.reserve(_indices[indexIndex].size());

    for (tIndex::iterator pos = _indices[indexIndex].begin();
      pos != _indices[indexIndex].end(); ++pos)
    {
        col.push_back(operator()(pos->second, colIndex));
    }

    }
    catch (NotFoundException& notFound)
    {
        notFound.AppendMessage("Column not found",
          "ISTable::GetColumn");
        throw;
    }

}

